
#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "access/tupmacs.h"
#include "utils/array.h"
#include "utils/lsyscache.h"

PG_MODULE_MAGIC;

Datum array_unnest(PG_FUNCTION_ARGS);
Datum array_agg_transfn(PG_FUNCTION_ARGS);
Datum array_agg_finalfn(PG_FUNCTION_ARGS);

static Datum ArrayCast(char *value, bool byval, int len);
static bool array_get_isnull(const bits8 *nullbitmap, int offset);

/*
 * Fetch array element at pointer, converted correctly to a Datum
 *
 * Caller must have handled case of NULL element
 */
static Datum
ArrayCast(char *value, bool byval, int len)
{
	return fetch_att(value, byval, len);
}

/*
 * Check whether a specific array element is NULL
 *
 * nullbitmap: pointer to array's null bitmap (NULL if none)
 * offset: 0-based linear element number of array element
 */
static bool
array_get_isnull(const bits8 *nullbitmap, int offset)
{
	if (nullbitmap == NULL)
		return false;	/* assume not null */
	if (nullbitmap[offset / 8] & (1 << (offset % 8)))
		return false;	/* not null */
	return true;
}


/*
 * UNNEST
 */

PG_FUNCTION_INFO_V1(array_unnest);
Datum
array_unnest(PG_FUNCTION_ARGS)
{
	typedef struct
	{
		ArrayType		*arr;
		int				 nextelem;
		int				 numelems;
		char			*elemdataptr;	/* this moves with nextelem */
		bits8			*arraynullsptr;	/* this does not */
		int16			 elmlen;
		bool			 elmbyval;
		char			 elmalign;
	} array_unnest_fctx;

	FuncCallContext *funcctx;
	array_unnest_fctx *fctx;
	MemoryContext oldcontext;

	/* stuff done only on the first call of the function */
	if (SRF_IS_FIRSTCALL())
	{
		ArrayType  *arr = PG_GETARG_ARRAYTYPE_P(0);

		/* create a function context for cross-call persistence */
		funcctx = SRF_FIRSTCALL_INIT();

		/*
		 * switch to memory context appropriate for multiple function calls
		 */
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

		/* allocate memory for user context */
		fctx = (array_unnest_fctx *) palloc(sizeof(array_unnest_fctx));

		/*
		 * Initialize state.  Note we assume that the originally passed
		 * array will stick around for the whole call series.
		 */
		fctx->arr = arr;
		fctx->nextelem = 0;
		fctx->numelems = ArrayGetNItems(ARR_NDIM(arr), ARR_DIMS(arr));

		fctx->elemdataptr = ARR_DATA_PTR(arr);
		fctx->arraynullsptr = ARR_NULLBITMAP(arr);

		get_typlenbyvalalign(ARR_ELEMTYPE(arr),
							 &fctx->elmlen,
							 &fctx->elmbyval,
							 &fctx->elmalign);

		funcctx->user_fctx = fctx;
		MemoryContextSwitchTo(oldcontext);
	}

	/* stuff done on every call of the function */
	funcctx = SRF_PERCALL_SETUP();
	fctx = funcctx->user_fctx;

	if (fctx->nextelem < fctx->numelems)
	{
		int	offset = fctx->nextelem++;
		Datum	elem;

		/*
		 * Check for NULL array element
		 */
		if (array_get_isnull(fctx->arraynullsptr, offset))
		{
			fcinfo->isnull = true;
			elem = (Datum) 0;
			/* elemdataptr does not move */
		}
		else
		{
			/*
			 * OK, get the element
			 */
			char   *ptr = fctx->elemdataptr;

			fcinfo->isnull = false;
			elem = ArrayCast(ptr, fctx->elmbyval, fctx->elmlen);

			/*
			 * Advance elemdataptr over it
			 */
			ptr = att_addlength_pointer(ptr, fctx->elmlen, ptr);
			ptr = (char *) att_align_nominal(ptr, fctx->elmalign);
			fctx->elemdataptr = ptr;
		}

		SRF_RETURN_NEXT(funcctx, elem);
	}
	else
	{
		/* do when there is no more left */
		SRF_RETURN_DONE(funcctx);
	}
}

/*
 * ARRAY_AGG
 */

PG_FUNCTION_INFO_V1(array_agg_transfn);
Datum
array_agg_transfn(PG_FUNCTION_ARGS)
{
	Oid arg1_typeid = get_fn_expr_argtype(fcinfo->flinfo, 1);

	if (arg1_typeid == InvalidOid)
		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
						errmsg("could not determine input data type")));

	/* cannot be called directly because of internal-type argument */
	Assert(fcinfo->context && IsA(fcinfo->context, AggState));

	PG_RETURN_POINTER(accumArrayResult(PG_ARGISNULL(0) ? NULL : (ArrayBuildState *) PG_GETARG_POINTER(0),
									   PG_ARGISNULL(1) ? (Datum) 0 : PG_GETARG_DATUM(1),
									   PG_ARGISNULL(1),
									   arg1_typeid,
									   ((AggState *) fcinfo->context)->aggcontext));
}

PG_FUNCTION_INFO_V1(array_agg_finalfn);
Datum
array_agg_finalfn(PG_FUNCTION_ARGS)
{
	/* cannot be called directly because of internal-type argument */
	Assert(fcinfo->context && IsA(fcinfo->context, AggState));

	if (PG_ARGISNULL(0))
		PG_RETURN_NULL();   /* returns null iff no input values */

	PG_RETURN_ARRAYTYPE_P(makeArrayResult((ArrayBuildState *) PG_GETARG_POINTER(0), CurrentMemoryContext));
}

